////////////////////////////////////////////////////////////////////////////////
//
//  File: OutputVtk.cpp
//
//  For more information, please see: http://www.nektar.info/
//
//  The MIT License
//
//  Copyright (c) 2006 Division of Applied Mathematics, Brown University (USA),
//  Department of Aeronautics, Imperial College London (UK), and Scientific
//  Computing and Imaging Institute, University of Utah (USA).
//
//  License for the specific language governing rights and limitations under
//  Permission is hereby granted, free of charge, to any person obtaining a
//  copy of this software and associated documentation files (the "Software"),
//  to deal in the Software without restriction, including without limitation
//  the rights to use, copy, modify, merge, publish, distribute, sublicense,
//  and/or sell copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included
//  in all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
//  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
//  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//  DEALINGS IN THE SOFTWARE.
//
//  Description: VTK file format output.
//
////////////////////////////////////////////////////////////////////////////////

#include <set>
#include <string>
#include <iomanip>
using namespace std;

#include <MultiRegions/ExpList.h>

#include "OutputVtk.h"

namespace Nektar
{
    namespace Utilities
    {
        ModuleKey OutputVtk::m_className =
            GetModuleFactory().RegisterCreatorFunction(
                ModuleKey(eOutputModule, "vtu"), OutputVtk::create,
                "Writes a VTU file.");

        OutputVtk::OutputVtk(FieldSharedPtr f) : OutputModule(f)
        {
            m_requireEquiSpaced = true;
        }

        OutputVtk::~OutputVtk()
        {
        }
        
        void OutputVtk::Process(po::variables_map &vm)
        {
            if(!m_f->m_exp.size()) // do nothing if no expansion defined
            {
                return; 
            }

            int i, j;
            if (m_f->m_verbose)
            {
                cout << "OutputVtk: Writing file..." << endl;
            }

            // Extract the output filename and extension
            string filename = m_config["outfile"].as<string>();
            
            // amend for parallel output if required 
            if(m_f->m_session->GetComm()->GetSize() != 1)
            {
                int    dot  = filename.find_last_of('.');
                string ext = filename.substr(dot,filename.length()-dot);
                string procId = "_P" + boost::lexical_cast<std::string>(m_f->m_session->GetComm()->GetRank());
                string start = filename.substr(0,dot);
                filename = start + procId + ext;
            }

            // Write solution.
            ofstream outfile(filename.c_str());
            WriteVtkHeader(outfile);
            
            // For each field write out field data for each expansion.
            for (i = 0; i < m_f->m_exp[0]->GetNumElmts(); ++i)
            {
                m_f->m_exp[0]->WriteVtkPieceHeader(outfile,i);
                // For this expansion write out each field.
                for (j = 0; j < m_f->m_fielddef[0]->m_fields.size(); ++j)
                {
                    WriteVtkPieceData(outfile, i, j);
                }
                WriteVtkPieceFooter(outfile, i);
            }
            WriteVtkFooter(outfile);
            cout << "Written file: " << filename << endl;
        }

        void OutputVtk::WriteVtkHeader(std::ostream &outfile)
        {
            outfile << "<?xml version=\"1.0\"?>" << endl;
            outfile << "<VTKFile type=\"UnstructuredGrid\" version=\"0.1\" "
                    << "byte_order=\"LittleEndian\">" << endl;
            outfile << "  <UnstructuredGrid>" << endl;
        }

        void OutputVtk::WriteVtkFooter(std::ostream &outfile)
        {
            outfile << "  </UnstructuredGrid>" << endl;
            outfile << "</VTKFile>" << endl;
        }

        /**
         * Writes out the header for a <PIECE> VTK XML segment describing the
         * geometric information which comprises this element. This includes
         * vertex coordinates for each quadrature point, vertex connectivity
         * information, cell types and cell offset data.
         *
         * @param   outfile     Output stream to write data to.
         * @param   expansion   Expansion to output data from.
         */
/*        void OutputVtk::WriteVtkPieceHeader(
            std::ofstream &outfile,
            int expansion)
        {
            const int dim = m_f->m_graph->GetMeshDimension();
            int i,j;
            int nquad0 = m_f->m_exp[0]->GetExp(expansion)->GetNumPoints(0);
            int nquad1 = m_f->m_exp[0]->GetExp(expansion)->GetNumPoints(1);
            int nquad2 = m_f->m_exp[0]->GetExp(expansion)->GetNumPoints(2);
            int ntot = nquad0*nquad1;
            int ntotminus = (nquad0-1)*(nquad1-1);

            Array<OneD,NekDouble> coords[3];
            coords[0] = Array<OneD,NekDouble>(ntot,0.0);
            coords[1] = Array<OneD,NekDouble>(ntot,0.0);
            coords[2] = Array<OneD,NekDouble>(ntot,0.0);
            m_f->m_exp[0]->GetExp(expansion)->GetCoords(coords[0],coords[1],coords[2]);

            outfile << "    <Piece NumberOfPoints=\""
                    << ntot << "\" NumberOfCells=\""
                    << ntotminus << "\">" << endl;
            outfile << "      <Points>" << endl;
            outfile << "        <DataArray type=\"Float64\" "
                    << "NumberOfComponents=\"3\" format=\"ascii\">" << endl;
            outfile << "          ";
            for (i = 0; i < ntot; ++i)
            {
                for (j = 0; j < 3; ++j)
                {
                    outfile << setprecision(8)     << scientific
                            << (float)coords[j][i] << " ";
                }
                outfile << endl;
            }
            outfile << endl;
            outfile << "        </DataArray>" << endl;
            outfile << "      </Points>" << endl;
            outfile << "      <Cells>" << endl;
            outfile << "        <DataArray type=\"Int32\" "
                    << "Name=\"connectivity\" format=\"ascii\">" << endl;
            switch (m_f->m_graph->GetMeshDimension())
            {
                case 1:
                    for (int i = 0; i < nquad0-1; ++i)
                    {
                        outfile << i << " " << i+1 << endl;
                    }
                    break;
                case 2:
                    for (int i = 0; i < nquad0-1; ++i)
                    {
                        for (int j = 0; j < nquad1-1; ++j)
                        {
                            outfile << j*nquad0 + i << " "
                                    << j*nquad0 + i + 1 << " "
                                    << (j+1)*nquad0 + i + 1 << " "
                                    << (j+1)*nquad0 + i << endl;
                        }
                    }
                    break;
                case 3:
                    for (int i = 0; i < nquad0-1; ++i)
                    {
                        for (int j = 0; j < nquad1-1; ++j)
                        {
                            for (int k = 0; k < nquad2-1; ++k)
                            {
                                outfile << k*nquad0*nquad1 + j*nquad0 + i << " "
                                        << k*nquad0*nquad1 + j*nquad0 + i + 1 << " "
                                        << k*nquad0*nquad1 + (j+1)*nquad0 + i + 1 << " "
                                        << k*nquad0*nquad1 + (j+1)*nquad0 + i << " "
                                        << (k+1)*nquad0*nquad1 + j*nquad0 + i << " "
                                        << (k+1)*nquad0*nquad1 + j*nquad0 + i + 1 << " "
                                        << (k+1)*nquad0*nquad1 + (j+1)*nquad0 + i + 1 << " "
                                        << (k+1)*nquad0*nquad1 + (j+1)*nquad0 + i << " " << endl;
                            }
                        }
                    }
                    break;
            }
            outfile << endl;
            outfile << "        </DataArray>" << endl;
            outfile << "        <DataArray type=\"Int32\" "
                    << "Name=\"offsets\" format=\"ascii\">" << endl;
            int shapeid;
            switch (m_f->m_graph->GetMeshDimension())
            {
                case 1:
                    for (int i = 0; i < nquad0-1; ++i)
                    {
                        outfile << i << " " << i+1 << endl;
                    }
                    shapeid=3;
                    break;
                case 2:
                    for (int i = 0; i < ntotminus; ++i)
                    {
                        outfile << i*4+4 << " ";
                    }
                    shapeid=9;
                    break;
                case 3:
                    for (int i = 0; i < ntotminus; ++i)
                    {
                        outfile << i*8+8 << " ";
                    }
                    shapeid=12;
                    break;
            }
            outfile << endl;
            outfile << "        </DataArray>" << endl;
            outfile << "        <DataArray type=\"UInt8\" "
                    << "Name=\"types\" format=\"ascii\">" << endl;
            for (i = 0; i < ntotminus; ++i)
            {
                outfile << shapeid << " ";
            }
            outfile << endl;
            outfile << "        </DataArray>" << endl;
            outfile << "      </Cells>" << endl;
            outfile << "      <PointData>" << endl;
        }*/

        void OutputVtk::WriteVtkPieceFooter(std::ofstream &outfile, int expansion)
        {
            outfile << "      </PointData>" << endl;
            outfile << "    </Piece>" << endl;
        }

        void OutputVtk::WriteVtkPieceData(std::ofstream &outfile, int expansion,
                                        int var)
        {
            int i;
            int nq = m_f->m_exp[var]->GetExp(expansion)->GetTotPoints();

            // printing the fields of that zone
            outfile << "        <DataArray type=\"Float32\" Name=\""
                    << m_f->m_fielddef[0]->m_fields[var] << "\">" << endl;
            outfile << "          ";
            const Array<OneD, NekDouble> phys = m_f->m_exp[var]->GetPhys() +
                                    m_f->m_exp[var]->GetPhys_Offset(expansion);

            for(i = 0; i < nq; ++i)
            {
                outfile << (fabs(phys[i]) < NekConstants::kNekZeroTol ? 0 : phys[i]) << " ";
            }

            outfile << endl;
            outfile << "        </DataArray>" << endl;
        }

    }
}
